# Copyright (c) 2021 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib.api import converters
from neutron_lib.api.definitions import qos as qos_apidef
from neutron_lib.api.definitions import qos_bw_minimum_ingress
from neutron_lib import constants
from neutron_lib.db import constants as db_const
from neutron_lib.services.qos import constants as qos_const


ALIAS = 'qos-pps'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'QoS Rule Type Packet per Second Extension'
API_PREFIX = ''
DESCRIPTION = 'Add QoS Rule Type Packet per Second'
UPDATED_TIMESTAMP = '2021-05-12T10:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {}
PACKET_RATE_LIMIT_RULES = 'packet_rate_limit_rules'
RESOURCE_NAME = 'packet_rate_limit_rule'
SUB_RES_ATTR_MAP = copy.deepcopy(
    qos_apidef.SUB_RESOURCE_ATTRIBUTE_MAP)
SUB_RES_ATTR_MAP.update({
    PACKET_RATE_LIMIT_RULES: {
        'parent': qos_apidef._PARENT,
        'parameters': dict(
            qos_apidef._QOS_RULE_COMMON_FIELDS,
            **{qos_const.MAX_KPPS: {
                'allow_post': True, 'allow_put': True,
                'convert_to': converters.convert_to_int,
                'is_visible': True,
                'is_filter': True,
                'is_sort_key': True,
                'validate': {
                    'type:range': [0, db_const.DB_INTEGER_MAX_VALUE]}
              },
              qos_const.MAX_BURST_KPPS: {
                'allow_post': True, 'allow_put': True,
                'is_visible': True, 'default': 0,
                'is_filter': True,
                'is_sort_key': True,
                'convert_to': converters.convert_to_int,
                'validate': {
                    'type:range': [0, db_const.DB_INTEGER_MAX_VALUE]}
              },
              qos_const.DIRECTION: {
                'allow_post': True,
                'allow_put': True,
                'is_visible': True,
                'is_filter': True,
                'is_sort_key': True,
                'default': constants.EGRESS_DIRECTION,
                'validate': {
                    'type:values': constants.VALID_DIRECTIONS}
              }
            }),
    }
})
SUB_RES_ATTR_MAP.update(
    qos_bw_minimum_ingress.SUB_RESOURCE_ATTRIBUTE_MAP)
SUB_RESOURCE_ATTRIBUTE_MAP = SUB_RES_ATTR_MAP
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [qos_apidef.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import reflection

from neutron_lib._i18n import _
from neutron_lib import exceptions


class NeutronObjectUpdateForbidden(exceptions.NeutronException):
    message = _("Unable to update the following object fields: %(fields)s")


class NeutronDbObjectDuplicateEntry(exceptions.Conflict):
    """Duplicate entry at unique column error.

    Raised when made an attempt to write to a unique column the same entry as
    existing one. :attr: `columns` available on an instance of the exception
    and could be used at error handling::

       try:
           obj_ref.save()
       except NeutronDbObjectDuplicateEntry as e:
           if 'colname' in e.columns:
               # Handle error.
    """
    message = _("Failed to create a duplicate %(object_type)s: "
                "for attribute(s) %(attributes)s with value(s) %(values)s")

    def __init__(self, object_class, db_exception):
        super().__init__(
            object_type=reflection.get_class_name(object_class,
                                                  fully_qualified=False),
            attributes=db_exception.columns,
            values=db_exception.value)
        self.columns = db_exception.columns
        self.value = db_exception.value


class NeutronDbObjectNotFoundByModel(exceptions.NotFound):
    message = _("NeutronDbObject not found by model %(model)s.")


class NeutronPrimaryKeyMissing(exceptions.BadRequest):
    message = _("For class %(object_type)s missing primary keys: "
                "%(missing_keys)s")

    def __init__(self, object_class, missing_keys):
        super().__init__(
            object_type=reflection.get_class_name(object_class,
                                                  fully_qualified=False),
            missing_keys=missing_keys
        )


class NeutronRangeConstrainedIntegerInvalidLimit(exceptions.NeutronException):
    message = _("Incorrect range limits specified: "
                "start = %(start)s, end = %(end)s")


class NeutronSyntheticFieldMultipleForeignKeys(exceptions.NeutronException):
    message = _("Synthetic field %(field)s shouldn't have more than one "
                "foreign key")


class NeutronSyntheticFieldsForeignKeysNotFound(exceptions.NeutronException):
    message = _("%(child)s does not define a foreign key for %(parent)s")


class NeutronObjectValidatorException(exceptions.NeutronException):
    message = _("Synthetic field(s) %(fields)s undefined, misspelled, or "
                "otherwise invalid")



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import dns

# The alias of the extension.
ALIAS = 'dns-integration-domain-keywords'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map (mandatory).
IS_SHIM_EXTENSION = True

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension (mandatory).
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension (mandatory).
NAME = 'DNS domain names with keywords allowed'

# A prefix for API resources. An empty prefix means that the API is going
# to be exposed at the v2/ level as any other core resource (mandatory).
API_PREFIX = ''

# The description of the extension (mandatory).
DESCRIPTION = ("Allows to use keywords like <project_id>, <project_name>, "
               "<user_id> and <user_name> as DNS domain name")

# A timestamp of when the extension was introduced (mandatory).
UPDATED_TIMESTAMP = "2020-06-15T18:00:00-00:00"

# The resource attribute map for the extension. It is effectively the
# bulk of the API contract alongside ACTION_MAP (mandatory).
RESOURCE_ATTRIBUTE_MAP = {}

# The subresource attribute map for the extension. It adds child resources
# to main extension's resource. The subresource map must have a parent and
# a parameters entry. If an extension does not need such a map, None can
# be specified (mandatory). For example:
SUB_RESOURCE_ATTRIBUTE_MAP = {}

# The action map: it associates verbs with methods to be performed on
# the API resource (mandatory).
ACTION_MAP = {}

# The action status: it associates response statuses with methods to be
# performed on the API resource (mandatory).
ACTION_STATUS = {}

# The list of required extensions (mandatory).
REQUIRED_EXTENSIONS = [dns.ALIAS]

# The list of optional extensions (mandatory).
OPTIONAL_EXTENSIONS = []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields as obj_fields


STANDARD_ATTRIBUTES = {
    'revision_number': obj_fields.IntegerField(),
    'description': obj_fields.StringField(nullable=True),
    'created_at': obj_fields.DateTimeField(nullable=True, tzinfo_aware=False),
    'updated_at': obj_fields.DateTimeField(nullable=True, tzinfo_aware=False),
}


def add_standard_attributes(cls):
    """Add standard attributes to the said class.

    :param cls: The class to add standard attributes to.
    :returns: None.
    """
    # Don't use parent's fields in case child class doesn't create
    # its own instance of list
    cls.fields = cls.fields.copy()
    cls.fields.update(STANDARD_ATTRIBUTES)
    # those fields are updated by sqlalchemy itself
    cls.fields_no_update += ('created_at', 'updated_at')
    # revision numbers are managed by service plugin and are bumped
    # automatically; consumers should not bump them explicitly
    cls.fields_no_update.append('revision_number')


def add_tag_filter_names(cls):
    """Add tag filter names to the said class.

    :param cls: The class to add tag filter names to.
    :returns: None.
    """
    cls.add_extra_filter_name("tags")
    cls.add_extra_filter_name("not-tags")
    cls.add_extra_filter_name("tags-any")
    cls.add_extra_filter_name("not-tags-any")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib.util
import os

from neutron_lib.api import definitions
from neutron_lib.api.definitions import base
from neutron_lib.api import validators
from neutron_lib import constants
from neutron_lib.tests import _base as test_base


def assert_bool(tester, attribute, attribute_dict, keyword, value):
    tester.assertIsInstance(
        value, bool,
        '%s must be a boolean for %s.' % (keyword, attribute))


def assert_converter(tester, attribute, attribute_dict, keyword, value):
    if ('default' not in attribute_dict or
            attribute_dict['default'] is constants.ATTR_NOT_SPECIFIED or
            attribute_dict.get(constants.DICT_POPULATE_DEFAULTS)):
        return
    try:
        attribute_dict['convert_to'](attribute_dict['default'])
    except KeyError:
        try:
            attribute_dict['convert_list_to'](attribute_dict['default'])
        except KeyError:
            if validators.is_attr_set(value) and not isinstance(
                    value, (str, list)):
                tester.fail("Default value '%s' cannot be converted for "
                            "attribute %s." % (value, attribute))


def assert_true(tester, attribute, attribute_dict, keyword, value):
    tester.assertTrue(
        value, '%s must be True for %s.' % (keyword, attribute))


def assert_validator(tester, attribute, attribute_dict, keyword, value):
    tester.assertIn(list(value)[0], validators.validators,
                    '%s is not a known validator for %s.' % (value, attribute))


ASSERT_FUNCTIONS = {
    'allow_post': assert_bool,
    'allow_put': assert_bool,
    'convert_to': assert_converter,
    'convert_list_to': assert_converter,
    'default': assert_converter,
    'enforce_policy': assert_bool,
    'is_filter': assert_bool,
    'is_sort_key': assert_bool,
    'is_visible': assert_bool,
    'primary_key': assert_true,
    'required_by_policy': assert_bool,
    'validate': assert_validator,
    'default_overrides_none': assert_bool,
    'dict_populate_defaults': assert_bool,
}


class DefinitionBaseTestCase(test_base.BaseTestCase):

    extension_module = None
    extension_resources = ()
    extension_subresources = ()
    extension_attributes = ()

    def setUp(self):
        super(DefinitionBaseTestCase, self).setUp()
        if not self.extension_module:
            self.fail("Missing extension module definition.")
        self.alias = self.extension_module.ALIAS
        self.is_shim_extension = self.extension_module.IS_SHIM_EXTENSION
        self.is_standard_attr_extension = (
            self.extension_module.IS_STANDARD_ATTR_EXTENSION)
        self.name = self.extension_module.NAME
        self.description = self.extension_module.DESCRIPTION
        self.resource_map = self.extension_module.RESOURCE_ATTRIBUTE_MAP
        self.subresource_map = self.extension_module.SUB_RESOURCE_ATTRIBUTE_MAP
        self.action_map = self.extension_module.ACTION_MAP
        self.action_status = self.extension_module.ACTION_STATUS
        self.required_extensions = self.extension_module.REQUIRED_EXTENSIONS
        self.optional_extensions = self.extension_module.OPTIONAL_EXTENSIONS

    def test_shim_extension(self):
        if self.is_shim_extension is True:
            self.assertFalse(self.extension_resources)
            self.assertFalse(self.extension_attributes)
            self.assertFalse(self.resource_map)
            self.assertFalse(self.action_map)
            self.assertFalse(self.action_status)

    def test_is_standard_attr_extension(self):
        if self.is_standard_attr_extension:
            self.assertIn('standard-attr-', self.alias)
        else:
            self.skipTest('API definition is not related to standardattr.')

    def test_resource_map(self):
        if (not self.resource_map and not self.subresource_map and
                not self.is_shim_extension and not self.action_map):
            self.fail('Missing resource map, subresource map, '
                      'and action map, what is this extension doing?')
        elif self.is_shim_extension:
            self.skipTest('Shim extension with no API changes.')

        for resource in self.resource_map:
            self.assertIn(
                resource, base.KNOWN_RESOURCES + self.extension_resources,
                'Resource is unknown, check for typos.')
            self.assertParams(self.resource_map[resource])

    def assertParams(self, resource):
        for attribute in resource.keys():
            self.assertIn(
                attribute,
                base.KNOWN_ATTRIBUTES + self.extension_attributes,
                'Attribute is unknown, check for typos.')
            for keyword in resource[attribute]:
                self.assertIn(keyword, base.KNOWN_KEYWORDS,
                              'Keyword is unknown, check for typos.')
                value = resource[attribute][keyword]
                assert_f = ASSERT_FUNCTIONS[keyword]
                assert_f(self, attribute,
                         resource[attribute],
                         keyword, value)

    def _assert_subresource(self, subresource):
        self.assertIn(
            self.subresource_map[subresource]['parent']['collection_name'],
            base.KNOWN_RESOURCES + self.extension_resources,
            'Sub-resource parent is unknown, check for typos.')
        self.assertIn('member_name',
                      self.subresource_map[subresource]['parent'],
                      'Incorrect parent definition, check for typos.')
        self.assertParams(self.subresource_map[subresource]['parameters'])

    def test_subresource_map(self):
        if not self.subresource_map:
            self.skipTest('API extension has no subresource map.')
        for subresource in self.subresource_map:
            self.assertIn(
                subresource,
                self.extension_subresources + base.KNOWN_RESOURCES,
                'Sub-resource is unknown, check for typos.')
            sub_attrmap = self.subresource_map[subresource]
            if 'parent' in sub_attrmap:
                self.assertEqual(2, len(sub_attrmap.keys()))
                self.assertIn('parent', sub_attrmap)
                self.assertIn('parameters', sub_attrmap)
                self._assert_subresource(subresource)
            else:
                self.assertEqual(
                    ['parameters'], [p for p in sub_attrmap.keys()],
                    'When extending sub-resources only use the parameters '
                    'keyword')
                self.assertParams(sub_attrmap['parameters'])

    def test_action_map(self):
        self.assertIsInstance(self.action_map, dict)
        if not self.action_map:
            self.skipTest('API definition has no action map.')

        for key in self.action_map:
            for action in self.action_map[key].values():
                self.assertIn(action, base.KNOWN_HTTP_ACTIONS,
                              'HTTP verb is unknown, check for typos.')

    def test_action_status(self):
        if not self.action_status:
            self.skipTest('API definition has no action status.')

        for status in self.action_status.values():
            self.assertIn(status, base.KNOWN_ACTION_STATUSES,
                          'HTTP status is unknown, check for typos.')

    def test_required_extensions(self):
        self.assertIsInstance(self.required_extensions, list)
        if not self.required_extensions:
            self.skipTest('API definition has no required extensions.')

        for ext in self.required_extensions:
            self.assertIn(ext, base.KNOWN_EXTENSIONS,
                          'Required extension is unknown, check for typos.')

    def test_optional_extensions(self):
        self.assertIsInstance(self.optional_extensions, list)
        if not self.optional_extensions:
            self.skipTest('API definition has no optional extensions.')

        for ext in self.optional_extensions:
            self.assertIn(ext, base.KNOWN_EXTENSIONS,
                          'Optional extension is unknown, check for typos.')

    def _load_module(self, name, path):
        module_spec = importlib.util.spec_from_file_location(
            name, path
        )
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        return module

    def test_all_api_definitions_list(self):
        # ensure _ALL_API_DEFINITIONS contains all public api-defs
        ext_aliases = []
        api_def_path = 'neutron_lib/api/definitions'
        for f in sorted(os.listdir(api_def_path)):
            mod_name, file_ext = os.path.splitext(os.path.split(f)[-1])
            ext_path = os.path.join(api_def_path, f)
            if file_ext.lower() == '.py' and not mod_name.startswith('_'):
                mod = self._load_module(mod_name, ext_path)
                ext_alias = getattr(mod, 'ALIAS', None)
                if not ext_alias:
                    continue
                ext_aliases.append(ext_alias)

        self.assertEqual(sorted(ext_aliases),
                         sorted([d.ALIAS for d in
                                 definitions._ALL_API_DEFINITIONS]))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from neutron_lib import constants

from neutron_lib.tests import _base as base


class TestNeutronLib(base.BaseTestCase):

    def test_sentinel_constant(self):
        foo = constants.Sentinel()
        bar = copy.deepcopy(foo)
        self.assertEqual(id(foo), id(bar))

    def test_sentinel_copy(self):
        singleton = constants.Sentinel()
        self.assertEqual(copy.deepcopy(singleton), copy.copy(singleton))



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import flowclassifier as fc_api
from neutron_lib import constants
from neutron_lib.db import constants as db_const

# The alias of the extension.
ALIAS = 'sfc'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'Service Function Chaining'

# The description of the extension.
DESCRIPTION = "Provides support for Service Function Chaining"

# A timestamp of when the extension was last updated.
UPDATED_TIMESTAMP = "2015-10-05T10:00:00-00:00"

API_PREFIX = '/sfc'

# Service Function Chaining constants
PORT_PAIRS = 'port_pairs'
PORT_PAIR_GROUPS = 'port_pair_groups'
PORT_CHAINS = 'port_chains'
MAX_CHAIN_ID = 65535

MPLS = 'mpls'
NSH = 'nsh'

PPG_N_TUPLE_DICT_SPEC = {
    'source_ip_prefix': {
        'default': None,
        'type:subnet_or_none': None
    },
    'destination_ip_prefix': {
        'default': None,
        'type:subnet_or_none': None
    },
    'source_port_range_min': {
        'default': None,
        'convert_to': converters.convert_to_int_if_not_none,
        'type:range_or_none': [0, constants.PORT_MAX]
    },
    'source_port_range_max': {
        'default': None,
        'convert_to': converters.convert_to_int_if_not_none,
        'type:range_or_none': [0, constants.PORT_MAX]
    },
    'destination_port_range_min': {
        'default': None,
        'convert_to': converters.convert_to_int_if_not_none,
        'type:range_or_none': [0, constants.PORT_MAX]
    },
    'destination_port_range_max': {
        'default': None,
        'convert_to': converters.convert_to_int_if_not_none,
        'type:range_or_none': [0, constants.PORT_MAX]
    }
}

# The resource attribute map for the extension.
RESOURCE_ATTRIBUTE_MAP = {
    PORT_PAIRS: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None},
            'primary_key': True
        },
        'name': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
        },
        'description': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
        },
        'tenant_id': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'required_by_policy': True
        },
        'ingress': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None}
        },
        'egress': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None}
        },
        'service_function_parameters': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'dict_populate_defaults': True,
            'validate': {
                'type:dict': {
                    'correlation': {
                        'type:values': [None, MPLS, NSH],
                        'default': None
                    },
                    'weight': {
                        'type:non_negative': None,
                        'default': 1,
                        'convert_to': converters.convert_to_int
                    }
                }
            }
        }
    },
    PORT_PAIR_GROUPS: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None},
            'primary_key': True},
        'group_id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True
        },
        'name': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
        },
        'description': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
        },
        'tenant_id': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'required_by_policy': True
        },
        'port_pairs': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': None,
            'validate': {'type:uuid_list': None},
            'convert_to': converters.convert_none_to_empty_list
        },
        'port_pair_group_parameters': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'dict_populate_defaults': True,
            'validate': {
                'type:dict': {
                    'lb_fields': {
                        'type:list_of_regex_or_none':
                            'eth|ip|tcp|udp)_(src|dst)',
                        'default': []
                    },
                    'ppg_n_tuple_mapping': {
                        'dict_populate_defaults': True,
                        'type:dict': {
                            'ingress_n_tuple': {
                                'type:dict': PPG_N_TUPLE_DICT_SPEC,
                                'default': {}
                            },
                            'egress_n_tuple': {
                                'type:dict': PPG_N_TUPLE_DICT_SPEC,
                                'default': {}
                            }
                        }
                    }
                }
            }
        }
    },
    PORT_CHAINS: {
        'id': {
            'allow_post': False, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:uuid': None},
            'primary_key': True
        },
        'chain_id': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True, 'default': 0,
            'validate': {'type:range': (0, MAX_CHAIN_ID)},
            'convert_to': converters.convert_to_int
        },
        'name': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
        },
        'description': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': '',
            'validate': {'type:string': db_const.DESCRIPTION_FIELD_SIZE},
        },
        'tenant_id': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'required_by_policy': True
        },
        'port_pair_groups': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True,
            'validate': {'type:uuid_list_non_empty': None},
            'default': None
        },
        'flow_classifiers': {
            'allow_post': True, 'allow_put': True,
            'is_visible': True, 'default': None,
            'validate': {'type:uuid_list': None},
            'convert_to': converters.convert_to_list
        },
        'chain_parameters': {
            'allow_post': True, 'allow_put': False,
            'is_visible': True,
            'dict_populate_defaults': True,
            'validate': {
                'type:dict': {
                    'correlation': {
                        'type:values': [MPLS, NSH],
                        'default': MPLS
                    },
                    'symmetric': {
                        'type:boolean': None,
                        'default': False,
                        'convert_to': converters.convert_to_boolean
                    }
                }
            },
        }
    }
}

# The subresource attribute map for the extension. It adds child resources
# to main extension's resource. The subresource map must have a parent and
# a parameters entry. If an extension does not need such a map, None can
# be specified (mandatory).
SUB_RESOURCE_ATTRIBUTE_MAP = None

# The action map.
ACTION_MAP = {
}

ACTION_STATUS = {
}
# The list of required extensions.
REQUIRED_EXTENSIONS = [
    fc_api.ALIAS,
]

# The list of optional extensions.
OPTIONAL_EXTENSIONS = [
]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import bgp
from neutron_lib import constants


ALIAS = 'bgp_4byte_asn'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = "BGP 4-byte AS numbers"
DESCRIPTION = "Support bgp 4-byte AS number"
UPDATED_TIMESTAMP = '2017-09-07T00:00:00-00:00'


RESOURCE_ATTRIBUTE_MAP = {
    'bgp-speakers': {
        'local_as': {'allow_post': True, 'allow_put': False,
                     'validate': {'type:range': (constants.MIN_ASNUM,
                                                 constants.MAX_4BYTE_ASNUM)},
                     'is_visible': True, 'default': None,
                     'required_by_policy': False,
                     'enforce_policy': False}
    },
    'bgp-peers': {
        'remote_as': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:range': (constants.MIN_ASNUM,
                                                  constants.MAX_4BYTE_ASNUM)},
                      'is_visible': True, 'default': None,
                      'required_by_policy': False,
                      'enforce_policy': False}
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = [bgp.ALIAS]
OPTIONAL_EXTENSIONS = []



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib._i18n import _
from neutron_lib import constants


_UNSET = constants.Sentinel()


def is_extension_supported(plugin, alias):
    """Validate that the extension is supported.

    :param plugin: The plugin class.
    :param alias: The alias to check.
    :returns: True if the alias is supported else False.
    """
    return alias in getattr(plugin, "supported_extension_aliases", [])


class ExtensionDescriptor(object, metaclass=abc.ABCMeta):
    """Base class that defines the contract for extensions."""

    @abc.abstractmethod
    def get_name(self):
        """The name of the extension.

        e.g. 'Fox In Socks'
        """

    @abc.abstractmethod
    def get_alias(self):
        """The alias for the extension.

        e.g. 'FOXNSOX'
        """

    @abc.abstractmethod
    def get_description(self):
        """Friendly description for the extension.

        e.g. 'The Fox In Socks Extension'
        """

    @abc.abstractmethod
    def get_updated(self):
        """The timestamp when the extension was last updated.

        e.g. '2011-01-22T13:25:27-06:00'
        """
        # NOTE(justinsb): Not sure of the purpose of this is, vs the XML NS

    def get_resources(self):
        """List of extensions.ResourceExtension extension objects.

        Resources define new nouns, and are accessible through URLs.
        """
        return []

    def get_actions(self):
        """List of extensions.ActionExtension extension objects.

        Actions are verbs callable from the API.
        """
        return []

    def get_request_extensions(self):
        """List of extensions.RequestExtension extension objects.

        Request extensions are used to handle custom request data.
        """
        return []

    def get_extended_resources(self, version):
        """Retrieve extended resources or attributes for core resources.

        Extended attributes are implemented by a core plugin similarly
        to the attributes defined in the core, and can appear in
        request and response messages. Their names are scoped with the
        extension's prefix. The core API version is passed to this
        function, which must return a
        map[<resource_name>][<attribute_name>][<attribute_property>]
        specifying the extended resource attribute properties required
        by that API version.
        Extension can add resources and their attr definitions too.
        The returned map can be integrated into RESOURCE_ATTRIBUTE_MAP.
        """
        return {}

    def get_plugin_interface(self):
        """Returns an abstract class which defines contract for the plugin.

        The abstract class should inherit from
        neutron_lib.services.base.ServicePluginBase.
        Methods in this abstract class should be decorated as abstractmethod
        """

    def get_required_extensions(self):
        """Return list of extensions required for processing this descriptor.

        Without these extensions present in a neutron deployment, the
        introduced extension cannot load or function properly.
        """
        return []

    def get_optional_extensions(self):
        """Returns a list of optionally required extensions.

        Unlike get_required_extensions. This will not fail the loading of
        the extension if one of these extensions is not present. This is
        useful for an extension that extends multiple resources across
        other extensions that should still work for the remaining extensions
        when one is missing.
        """
        return []

    @classmethod
    def update_attributes_map(cls, extended_attributes,
                              extension_attrs_map=None):
        """Update attributes map for this extension.

        This is default method for extending an extension's attributes map.
        An extension can use this method and supplying its own resource
        attribute map in extension_attrs_map argument to extend all its
        attributes that needs to be extended.
        If an extension does not implement update_attributes_map, the method
        does nothing and just return.
        """
        if not extension_attrs_map:
            return

        for resource, attrs in extension_attrs_map.items():
            extended_attrs = extended_attributes.get(resource)
            if extended_attrs:
                attrs.update(extended_attrs)

    def get_pecan_resources(self):
        """List of PecanResourceExtension extension objects.

        Resources define new nouns, and are accessible through URLs.
        The controllers associated with each instance of
        extensions.ResourceExtension should be a subclass of
        neutron.pecan_wsgi.controllers.utils.NeutronPecanController.
        If a resource is defined in both get_resources and get_pecan_resources,
        the resource defined in get_pecan_resources will take precedence.
        """
        return []


class APIExtensionDescriptor(ExtensionDescriptor):
    """Base class that defines the contract for extensions.

    Concrete implementations of API extensions should first provide
    an API definition in neutron_lib.api.definitions. The API
    definition module (object reference) can then be specified as a
    class level attribute on the concrete extension.

    For example::

        from neutron_lib.api.definitions import provider_net
        from neutron_lib.api import extensions


        class Providernet(extensions.APIExtensionDescriptor):
            api_definition = provider_net
            # nothing else needed if default behavior is acceptable


    If extension implementations need to override the default behavior of
    this class they can override the respective method directly.
    """
    api_definition = _UNSET

    @classmethod
    def _assert_api_definition(cls, attr=None):
        if cls.api_definition == _UNSET:
            raise NotImplementedError(
                _("Extension module API definition not set."))
        if attr and getattr(cls.api_definition, attr, _UNSET) == _UNSET:
            raise NotImplementedError(_("Extension module API definition "
                                        "does not define '%s'") % attr)

    @classmethod
    def get_name(cls):
        """The name of the API definition."""
        cls._assert_api_definition('NAME')
        return cls.api_definition.NAME

    @classmethod
    def get_alias(cls):
        """The alias for the API definition."""
        cls._assert_api_definition('ALIAS')
        return cls.api_definition.ALIAS

    @classmethod
    def get_description(cls):
        """Friendly description for the API definition."""
        cls._assert_api_definition('DESCRIPTION')
        return cls.api_definition.DESCRIPTION

    @classmethod
    def get_updated(cls):
        """The timestamp when the API definition was last updated."""
        cls._assert_api_definition('UPDATED_TIMESTAMP')
        return cls.api_definition.UPDATED_TIMESTAMP

    @classmethod
    def get_extended_resources(cls, version):
        """Retrieve the extended resource map for the API definition.

        :param version: The API version to retrieve the resource attribute
            map for.
        :returns: The extended resource map for the underlying API definition
            if the version is 2.0. The extended resource map returned includes
            both the API definition's RESOURCE_ATTRIBUTE_MAP and
            SUB_RESOURCE_ATTRIBUTE_MAP where applicable. If the version is
            not 2.0, an empty dict is returned.
        """
        if version == "2.0":
            cls._assert_api_definition('RESOURCE_ATTRIBUTE_MAP')
            cls._assert_api_definition('SUB_RESOURCE_ATTRIBUTE_MAP')
            # support api defs that use None for sub attr map
            sub_attrs = cls.api_definition.SUB_RESOURCE_ATTRIBUTE_MAP or {}
            return dict(
                list(cls.api_definition.RESOURCE_ATTRIBUTE_MAP.items()) +
                list(sub_attrs.items()))
        else:
            return {}

    @classmethod
    def get_required_extensions(cls):
        """Returns the API definition's required extensions."""
        cls._assert_api_definition('REQUIRED_EXTENSIONS')
        return cls.api_definition.REQUIRED_EXTENSIONS

    @classmethod
    def get_optional_extensions(cls):
        """Returns the API definition's optional extensions."""
        cls._assert_api_definition('OPTIONAL_EXTENSIONS')
        return cls.api_definition.OPTIONAL_EXTENSIONS

    @classmethod
    def update_attributes_map(cls, extended_attributes,
                              extension_attrs_map=None):
        """Update attributes map for this extension.

        Behaves like ExtensionDescriptor.update_attributes_map(), but
        if extension_attrs_map is not given the dict returned from
        self.get_extended_resources('2.0') is used.
        """
        if extension_attrs_map is None:
            extension_attrs_map = cls.get_extended_resources('2.0')
        super().update_attributes_map(
            extended_attributes, extension_attrs_map=extension_attrs_map)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test functions."""

import functools


def unstable_test(reason):
    """Decorator used to mark test as unstable, on failure test will be skipped

    :param reason: String used in explanation, for example, 'bug 123456'.
    """
    def decor(f):
        @functools.wraps(f)
        def inner(self, *args, **kwargs):
            try:
                return f(self, *args, **kwargs)
            except Exception as e:
                msg = ("%s was marked as unstable because of %s, "
                       "failure was: %s") % (self.id(), reason, e)
                raise self.skipTest(msg)
        return inner
    return decor



# Copyright 2022
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.agent.common import constants


def _replace_register(flow_params, register_number, register_value):
    """Replace value from flows to given register number

    'register_value' key in dictionary will be replaced by register number
    given by 'register_number'

    :param flow_params: Dictionary containing defined flows
    :param register_number: The number of register where value will be stored
    :param register_value: Key to be replaced by register number

    """
    try:
        reg_port = flow_params[register_value]
        del flow_params[register_value]
        flow_params['reg{:d}'.format(register_number)] = reg_port
    except KeyError:
        pass


def create_reg_numbers(flow_params):
    """Replace reg_(port|net) values with defined register numbers"""
    _replace_register(flow_params, constants.REG_PORT, constants.PORT_REG_NAME)
    _replace_register(flow_params, constants.REG_NET, constants.NET_REG_NAME)
    _replace_register(flow_params,
                      constants.REG_REMOTE_GROUP,
                      constants.REMOTE_GROUP_REG_NAME)
    _replace_register(flow_params,
                      constants.REG_MIN_BW,
                      constants.MIN_BW_REG_NAME)
    _replace_register(flow_params,
                      constants.REG_INGRESS_BW_LIMIT,
                      constants.INGRESS_BW_LIMIT_REG_NAME)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_config import cfg
from oslo_db.sqlalchemy import models
import sqlalchemy as sa
from sqlalchemy.ext import declarative
from sqlalchemy import orm

from neutron_lib.api import attributes
from neutron_lib import context
from neutron_lib.db import utils
from neutron_lib import exceptions as n_exc

from neutron_lib.tests import _base as base


class FakePort(declarative.declarative_base(cls=models.ModelBase)):
    __tablename__ = 'fakeports'
    port_id = sa.Column(sa.String(36), primary_key=True)
    name = sa.Column(sa.String(64))
    status = sa.Column(sa.String(16), nullable=False)


class FakeRouter(declarative.declarative_base(cls=models.ModelBase)):
    __tablename__ = 'fakerouters'
    router_id = sa.Column(sa.String(36), primary_key=True)
    gw_port_id = sa.Column(sa.String(36), sa.ForeignKey(FakePort.port_id))
    gw_port = orm.relationship(FakePort, lazy='joined')


class TestUtilsLegacyPolicies(base.BaseTestCase):

    def setUp(self):
        cfg.CONF.set_override('enforce_scope', False, group='oslo_policy')
        super().setUp()

    def test_get_sort_dirs(self):
        sorts = [(1, True), (2, False), (3, True)]
        self.assertEqual(['asc', 'desc', 'asc'],
                         utils.get_sort_dirs(sorts))

    def test_get_sort_dirs_reversed(self):
        sorts = [(1, True), (2, False), (3, True)]
        self.assertEqual(['desc', 'asc', 'desc'],
                         utils.get_sort_dirs(sorts, page_reverse=True))

    def test_get_and_validate_sort_keys(self):
        sorts = [('name', False), ('status', True)]
        self.assertEqual(['name', 'status'],
                         utils.get_and_validate_sort_keys(sorts, FakePort))

    def test_get_and_validate_sort_keys_bad_key_fails(self):
        sorts = [('master', True)]
        self.assertRaises(n_exc.BadRequest,
                          utils.get_and_validate_sort_keys, sorts, FakePort)

    def test_get_and_validate_sort_keys_by_relationship_fails(self):
        sorts = [('gw_port', True)]
        self.assertRaises(n_exc.BadRequest,
                          utils.get_and_validate_sort_keys, sorts, FakeRouter)

    def test_get_marker_obj(self):
        plugin = mock.Mock()
        plugin._get_myr.return_value = 'obj'
        obj = utils.get_marker_obj(plugin, 'ctx', 'myr', 10, mock.ANY)
        self.assertEqual('obj', obj)
        plugin._get_myr.assert_called_once_with('ctx', mock.ANY)

    def test_get_marker_obj_no_limit_and_marker(self):
        self.assertIsNone(utils.get_marker_obj(
            mock.Mock(), 'ctx', 'myr', 0, mock.ANY))
        self.assertIsNone(utils.get_marker_obj(
            mock.Mock(), 'ctx', 'myr', 10, None))

    @mock.patch.object(attributes, 'populate_project_info')
    def test_resource_fields(self, mock_populate):
        r = {
            'name': 'n',
            'id': '1',
            'desc': None
        }
        utils.resource_fields(r, ['name'])
        mock_populate.assert_called_once_with({'name': 'n'})

    def test_model_query_scope_is_project_admin(self):
        ctx = context.Context(
            project_id='some project',
            is_admin=True,
            is_advsvc=False)
        model = mock.Mock(project_id='project')

        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))

        # Ensure that project_id isn't mocked
        del model.project_id
        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))

    def test_model_query_scope_is_project_advsvc(self):
        ctx = context.Context(
            project_id='some project',
            is_admin=False,
            is_advsvc=True)
        model = mock.Mock(project_id='project')

        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))

        # Ensure that project_id isn't mocked
        del model.project_id
        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))

    def test_model_query_scope_is_project_regular_user(self):
        ctx = context.Context(
            project_id='some project',
            is_admin=False,
            is_advsvc=False)
        model = mock.Mock(project_id='project')

        self.assertTrue(
            utils.model_query_scope_is_project(ctx, model))

        # Ensure that project_id isn't mocked
        del model.project_id
        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))

    def test_model_query_scope_is_project_system_scope(self):
        ctx = context.Context(system_scope='all')
        model = mock.Mock(project_id='project')

        self.assertTrue(
            utils.model_query_scope_is_project(ctx, model))

        # Ensure that project_id isn't mocked
        del model.project_id
        self.assertFalse(
            utils.model_query_scope_is_project(ctx, model))


class TestUtilsWithScopeEnforcement(TestUtilsLegacyPolicies):

    def setUp(self):
        super().setUp()
        cfg.CONF.set_override('enforce_scope', True, group='oslo_policy')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import flowclassifier
from neutron_lib.tests.unit.api.definitions import base


class FlowClassifierDefinitionTestCase(base.DefinitionBaseTestCase):
    extension_module = flowclassifier
    extension_resources = (flowclassifier.COLLECTION_NAME,)
    extension_attributes = ('ethertype',
                            'protocol',
                            'source_port_range_min',
                            'source_port_range_max',
                            'destination_port_range_min',
                            'destination_port_range_max',
                            'source_ip_prefix',
                            'destination_ip_prefix',
                            'logical_destination_port',
                            'logical_source_port',
                            'l7_parameters')



# Copyright (c) 2021 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class QuotaDriverAPI(object, metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def get_default_quotas(context, resources, project_id):
        """Retrieve the default quotas for the list of resources and project.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resource keys.
        :param project_id: The ID of the project to return default quotas for.
        :return: dict from resource name to dict of name and limit
        """

    @staticmethod
    @abc.abstractmethod
    def get_project_quotas(context, resources, project_id):
        """Retrieve the quotas for the given list of resources and project

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resource keys.
        :param project_id: The ID of the project to return quotas for.
        :return: dict from resource name to dict of name and limit
        """

    @staticmethod
    @abc.abstractmethod
    def get_detailed_project_quotas(context, resources, project_id):
        """Retrieve detailed quotas for the given list of resources and project

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resource keys.
        :param project_id: The ID of the project to return quotas for.
        :return dict: mapping resource name in dict to its corresponding limit
                      used and reserved. Reserved currently returns default
                      value of 0
        """

    @staticmethod
    @abc.abstractmethod
    def delete_project_quota(context, project_id):
        """Delete the quota entries for a given project_id.

        After deletion, this project will use default quota values in conf.
        Raise a "not found" error if the quota for the given project was
        never defined.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to return quotas for.
        """

    @staticmethod
    @abc.abstractmethod
    def get_all_quotas(context, resources):
        """Given a list of resources, retrieve the quotas for the all tenants.

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resource keys.
        :return: quotas list of dict of project_id:, resourcekey1:
                 resourcekey2: ...
        """

    @staticmethod
    @abc.abstractmethod
    def update_quota_limit(context, project_id, resource, limit):
        """Update the quota limit for a resource in a project

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to update the quota.
        :param resource: the resource to update the quota.
        :param limit: new resource quota limit.
        """

    @staticmethod
    @abc.abstractmethod
    def make_reservation(context, project_id, resources, deltas, plugin):
        """Make multiple resource reservations for a given project

        :param context: The request context, for access checks.
        :param resources: A dictionary of the registered resource keys.
        :param project_id: The ID of the project to make the reservations for.
        :return: ``ReservationInfo`` object.
        """

    @staticmethod
    @abc.abstractmethod
    def commit_reservation(context, reservation_id):
        """Commit a reservation register

        :param context: The request context, for access checks.
        :param reservation_id: ID of the reservation register to commit.
        """

    @staticmethod
    @abc.abstractmethod
    def cancel_reservation(context, reservation_id):
        """Cancel a reservation register

        :param context: The request context, for access checks.
        :param reservation_id: ID of the reservation register to cancel.
        """

    @staticmethod
    @abc.abstractmethod
    def limit_check(context, project_id, resources, values):
        """Check simple quota limits.

        For limits--those quotas for which there is no usage
        synchronization function--this method checks that a set of
        proposed values are permitted by the limit restriction.

        If any of the proposed values is over the defined quota, an
        OverQuota exception will be raised with the sorted list of the
        resources which are too high.  Otherwise, the method returns
        nothing.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to make the reservations for.
        :param resources: A dictionary of the registered resource.
        :param values: A dictionary of the values to check against the
                       quota.
        """

    @staticmethod
    @abc.abstractmethod
    def get_resource_usage(context, project_id, resources, resource_name):
        """Return the resource current usage

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to make the reservations for.
        :param resources: A dictionary of the registered resources.
        :param resource_name: The name of the resource to retrieve the usage.
        :return: The current resource usage.
        """

    @staticmethod
    @abc.abstractmethod
    def quota_limit_check(context, project_id, resources, deltas):
        """Check the current resource usage against a set of deltas.

        This method will check if the provided resource deltas could be
        assigned depending on the current resource usage and the quota limits.
        If the resource deltas plus the resource usage fit under the quota
        limit, the method will pass. If not, a ``OverQuota`` will be raised.

        :param context: The request context, for access checks.
        :param project_id: The ID of the project to make the reservations for.
        :param resources: A dictionary of the registered resource.
        :param deltas: A dictionary of the values to check against the
                       quota limits.
        :return: None if passed; ``OverQuota`` if quota limits are exceeded,
                 ``InvalidQuotaValue`` if delta values are invalid.
        """

    @staticmethod
    @abc.abstractmethod
    def get_workers():
        """Return the quota driver workers to be spawned during initialization

        This method returns the quota driver workers that needs to be spawned
        during the plugin initialization. For example, ``DbQuotaNoLockDriver``
        requires a ``PeriodicWorker`` to clean up the expired reservations left
        in the database.

        :return: list of ``worker.BaseWorker`` or derived instances.
        """



# Copyright (c) 2016 NEC Technologies Ltd.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import network


L2_ADJACENCY = 'l2_adjacency'

ALIAS = L2_ADJACENCY
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'L2 Adjacency'
API_PREFIX = ''
DESCRIPTION = 'Display L2 Adjacency for Neutron Networks.'
UPDATED_TIMESTAMP = '2016-04-12T16:00:00-00:00'
RESOURCE_NAME = network.RESOURCE_NAME
COLLECTION_NAME = network.COLLECTION_NAME
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        L2_ADJACENCY: {
            'allow_post': False,
            'allow_put': False,
            'is_visible': True
        }
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


ALIAS = 'bgp_dragent_scheduler'
IS_SHIM_EXTENSION = True
IS_STANDARD_ATTR_EXTENSION = False
NAME = "BGP Dynamic Routing Agent Scheduler"
DESCRIPTION = "Schedules BgpSpeakers on BgpDrAgent"
UPDATED_TIMESTAMP = '2015-07-30T10:00:00-00:00'
BGP_DRINSTANCE = 'bgp-drinstance'
BGP_DRINSTANCES = BGP_DRINSTANCE + 's'
BGP_DRAGENT = 'bgp-dragent'
BGP_DRAGENTS = BGP_DRAGENT + 's'
RESOURCE_ATTRIBUTE_MAP = {}
SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port
from neutron_lib.api.definitions import trunk
from neutron_lib import constants


# The alias of the extension.
ALIAS = 'trunk-details'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'Trunk port details'

# The description of the extension.
DESCRIPTION = "Expose trunk port details"

# A timestamp of when the extension was last updated.
UPDATED_TIMESTAMP = "2016-01-01T10:00:00-00:00"
# TODO(armax): to be removed when review
# https://review.opendev.org/#/c/484058/ merges
TIMESTAMP = UPDATED_TIMESTAMP

# The name of the resource introduced or being extended.
RESOURCE_NAME = port.RESOURCE_NAME

# The plural for the resource introduced or being extended.
COLLECTION_NAME = port.COLLECTION_NAME

# The specific resources and/or attributes for the extension (optional).
TRUNK_DETAILS = 'trunk_details'

# The resource attribute map for the extension.
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {TRUNK_DETAILS: {'allow_post': False, 'allow_put': False,
                                      'default': constants.ATTR_NOT_SPECIFIED,
                                      'is_visible': True,
                                      'enforce_policy': True,
                                      'required_by_policy': True}},
}

# The subresource attribute map for the extension.
SUB_RESOURCE_ATTRIBUTE_MAP = None

# The action map.
ACTION_MAP = {}

# The action status.
ACTION_STATUS = {
}

# The list of required extensions.
REQUIRED_EXTENSIONS = [trunk.ALIAS]

# The list of optional extensions.
OPTIONAL_EXTENSIONS = []



# Copyright 2021 Huawei, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from neutron_lib.db import constants as db_const

RESOURCE_NAME = 'local_ip'
COLLECTION_NAME = RESOURCE_NAME + 's'

LOCAL_IP_ASSOCIATION = 'port_association'
LOCAL_IP_ASSOCIATIONS = LOCAL_IP_ASSOCIATION + 's'

ALIAS = RESOURCE_NAME
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Local IP'
DESCRIPTION = 'Support Node Local Virtual IP'
UPDATED_TIMESTAMP = '2021-07-26T10:00:00-00:00'

IP_MODE_TRANSLATE = 'translate'
IP_MODE_PASSTHROUGH = 'passthrough'
VALID_IP_MODES = [IP_MODE_TRANSLATE, IP_MODE_PASSTHROUGH]

RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'is_visible': True,
            'is_filter': True,
            'is_sort_key': True,
            'primary_key': True},
        'name': {
            'allow_post': True,
            'allow_put': True,
            'default': '',
            'validate': {'type:string': db_const.NAME_FIELD_SIZE},
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
        'description': {
            'allow_post': True,
            'allow_put': True,
            'default': '',
            'validate': {'type:string': db_const.LONG_DESCRIPTION_FIELD_SIZE},
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
        'project_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
            'required_by_policy': True,
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
        'local_port_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'default': constants.ATTR_NOT_SPECIFIED,
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
        'network_id': {
            'allow_post': True,
            'allow_put': False,
            'validate': {'type:uuid': None},
            'default': constants.ATTR_NOT_SPECIFIED,
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
        'local_ip_address': {
            'allow_post': True,
            'allow_put': False,
            'default': constants.ATTR_NOT_SPECIFIED,
            'validate': {'type:ip_address_or_none': None},
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True,
            'enforce_policy': True},
        'ip_mode': {
            'allow_post': True,
            'allow_put': False,
            'default': IP_MODE_TRANSLATE,
            'validate': {'type:values': VALID_IP_MODES},
            'is_filter': True,
            'is_sort_key': True,
            'is_visible': True},
    }
}

SUB_RESOURCE_ATTRIBUTE_MAP = {
    LOCAL_IP_ASSOCIATIONS: {
        'parent': {'collection_name': COLLECTION_NAME,
                   'member_name': RESOURCE_NAME},
        'parameters': {
            'local_ip_id': {
                'allow_post': False,
                'allow_put': False,
                'is_visible': True},
            'local_ip_address': {
                'allow_post': False,
                'allow_put': False,
                'is_filter': True,
                'is_sort_key': True,
                'is_visible': True},
            'fixed_port_id': {
                'allow_post': True,
                'allow_put': False,
                'validate': {'type:uuid': None},
                'is_filter': True,
                'is_sort_key': True,
                'is_visible': True},
            'fixed_ip': {
                'allow_post': True,
                'allow_put': False,
                'validate': {'type:ip_address_or_none': None},
                'is_filter': True,
                'is_sort_key': True,
                'is_visible': True,
                'default': None},
            'host': {
                'allow_post': False,
                'allow_put': False,
                'is_filter': True,
                'is_sort_key': True,
                'is_visible': True},
            'project_id': {
                'allow_post': True,
                'allow_put': False,
                'validate': {'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                'required_by_policy': True,
                'is_visible': False},
        }
    }
}

ACTION_MAP = {}

ACTION_STATUS = {}

REQUIRED_EXTENSIONS = []

OPTIONAL_EXTENSIONS = []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib._i18n import _
from neutron_lib import exceptions


# VPNaaS Exceptions
class VPNServiceNotFound(exceptions.NotFound):
    message = _("VPNService %(vpnservice_id)s could not be found")


class IPsecSiteConnectionNotFound(exceptions.NotFound):
    message = _("ipsec_site_connection %(ipsec_site_conn_id)s not found")


class IPsecSiteConnectionDpdIntervalValueError(exceptions.InvalidInput):
    message = _("ipsec_site_connection %(attr)s is "
                "equal to or less than dpd_interval")


class IPsecSiteConnectionMtuError(exceptions.InvalidInput):
    message = _("ipsec_site_connection MTU %(mtu)d is too small "
                "for ipv%(version)s")


class IPsecSiteConnectionPeerCidrError(exceptions.InvalidInput):
    message = _("ipsec_site_connection peer cidr %(peer_cidr)s is "
                "invalid CIDR")


class IKEPolicyNotFound(exceptions.NotFound):
    message = _("IKEPolicy %(ikepolicy_id)s could not be found")


class IPsecPolicyNotFound(exceptions.NotFound):
    message = _("IPsecPolicy %(ipsecpolicy_id)s could not be found")


class IKEPolicyInUse(exceptions.InUse):
    message = _("IKEPolicy %(ikepolicy_id)s is in use by existing "
                "IPsecSiteConnection and can't be updated or deleted")


class VPNServiceInUse(exceptions.InUse):
    message = _("VPNService %(vpnservice_id)s is still in use")


class SubnetInUseByVPNService(exceptions.InUse):
    message = _("Subnet %(subnet_id)s is used by VPNService %(vpnservice_id)s")


class SubnetInUseByEndpointGroup(exceptions.InUse):
    message = _("Subnet %(subnet_id)s is used by endpoint group %(group_id)s")


class SubnetInUseByIPsecSiteConnection(exceptions.InUse):
    message = _("Subnet %(subnet_id)s is used by ipsec site connection "
                "%(ipsec_site_connection_id)s")


class VPNStateInvalidToUpdate(exceptions.BadRequest):
    message = _("Invalid state %(state)s of vpnaas resource %(id)s "
                "for updating")


class IPsecPolicyInUse(exceptions.InUse):
    message = _("IPsecPolicy %(ipsecpolicy_id)s is in use by existing "
                "IPsecSiteConnection and can't be updated or deleted")


class DeviceDriverImportError(exceptions.NeutronException):
    message = _("Can not load driver :%(device_driver)s")


class SubnetIsNotConnectedToRouter(exceptions.BadRequest):
    message = _("Subnet %(subnet_id)s is not "
                "connected to Router %(router_id)s")


class RouterIsNotExternal(exceptions.BadRequest):
    message = _("Router %(router_id)s has no external network gateway set")


class VPNPeerAddressNotResolved(exceptions.InvalidInput):
    message = _("Peer address %(peer_address)s cannot be resolved")


class ExternalNetworkHasNoSubnet(exceptions.BadRequest):
    message = _("Router's %(router_id)s external network has "
                "no %(ip_version)s subnet")


# VPN Endpoint Group Exceptions
class VPNEndpointGroupNotFound(exceptions.NotFound):
    message = _("Endpoint group %(endpoint_group_id)s could not be found")


class InvalidEndpointInEndpointGroup(exceptions.InvalidInput):
    message = _("Endpoint '%(endpoint)s' is invalid for group "
                "type '%(group_type)s': %(why)s")


class MissingEndpointForEndpointGroup(exceptions.BadRequest):
    message = _("No endpoints specified for endpoint group '%(group)s'")


class NonExistingSubnetInEndpointGroup(exceptions.InvalidInput):
    message = _("Subnet %(subnet)s in endpoint group does not exist")


class MixedIPVersionsForIPSecEndpoints(exceptions.BadRequest):
    message = _("Endpoints in group %(group)s do not have the same IP "
                "version, as required for IPSec site-to-site connection")


class MixedIPVersionsForPeerCidrs(exceptions.BadRequest):
    message = _("Peer CIDRs do not have the same IP version, as required "
                "for IPSec site-to-site connection")


class MixedIPVersionsForIPSecConnection(exceptions.BadRequest):
    message = _("IP versions are not compatible between peer and local "
                "endpoints")


class InvalidEndpointGroup(exceptions.BadRequest):
    message = _("Endpoint group%(suffix)s %(which)s cannot be specified, "
                "when VPN Service has subnet specified")


class WrongEndpointGroupType(exceptions.BadRequest):
    message = _("Endpoint group %(which)s type is '%(group_type)s' and "
                "should be '%(expected)s'")


class PeerCidrsInvalid(exceptions.BadRequest):
    message = _("Peer CIDRs cannot be specified, when using endpoint "
                "groups")


class MissingPeerCidrs(exceptions.BadRequest):
    message = _("Missing peer CIDRs for IPsec site-to-site connection")


class MissingRequiredEndpointGroup(exceptions.BadRequest):
    message = _("Missing endpoint group%(suffix)s %(which)s for IPSec "
                "site-to-site connection")


class EndpointGroupInUse(exceptions.BadRequest):
    message = _("Endpoint group %(group_id)s is in use and cannot be deleted")


# VPN Flavors Exceptions
class FlavorsPluginNotLoaded(exceptions.NotFound):
    message = _("Flavors plugin not found")


class NoProviderFoundForFlavor(exceptions.NotFound):
    message = _("No service provider found for flavor %(flavor_id)s")


# IPsec Service Driver Validator Exceptions
class IpsecValidationFailure(exceptions.BadRequest):
    message = _("IPSec does not support %(resource)s attribute %(key)s "
                "with value '%(value)s'")


class IkeValidationFailure(exceptions.BadRequest):
    message = _("IKE does not support %(resource)s attribute %(key)s "
                "with value '%(value)s'")


# Cisco Csr Driver Exceptions
class CsrInternalError(exceptions.NeutronException):
    message = _("Fatal - %(reason)s")


# Cisco CSR Driver Validator Exceptions
class CsrValidationFailure(exceptions.BadRequest):
    message = _("Cisco CSR does not support %(resource)s attribute %(key)s "
                "with value '%(value)s'")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import random
import socket
from unittest import mock

from neutron_lib import constants
from neutron_lib.tests import _base as base
from neutron_lib.utils import net


class TestGetHostname(base.BaseTestCase):

    @mock.patch.object(socket, 'gethostname',
                       return_value='fake-host-name')
    def test_get_hostname(self, mock_gethostname):
        self.assertEqual('fake-host-name',
                         net.get_hostname())
        mock_gethostname.assert_called_once_with()


class TestGetRandomMac(base.BaseTestCase):

    @mock.patch.object(random, 'getrandbits', return_value=0xa2)
    def test_first_4_octets_unchanged(self, mock_rnd):
        mac = net.get_random_mac(['aa', 'bb', '00', 'dd', 'ee', 'ff'])
        self.assertEqual('aa:bb:00:dd:a2:a2', mac)
        mock_rnd.assert_called_with(8)

    @mock.patch.object(random, 'getrandbits', return_value=0xa2)
    def test_first_4th_octet_generated(self, mock_rnd):
        mac = net.get_random_mac(['aa', 'bb', 'cc', '00', 'ee', 'ff'])
        self.assertEqual('aa:bb:cc:a2:a2:a2', mac)
        mock_rnd.assert_called_with(8)


class TestRandomMacGenerator(base.BaseTestCase):

    def test_all_macs_generated(self):
        mac = ['aa', 'bb', 'cc', 'dd', 'ee', 'ff']
        generator = itertools.islice(net.random_mac_generator(mac), 70000)
        self.assertEqual(2**16, len(list(generator)))

    @mock.patch.object(random, 'getrandbits', return_value=0xa2)
    def test_first_generated_mac(self, mock_rnd):
        mac = ['aa', 'bb', 'cc', '00', 'ee', 'ff']
        generator = itertools.islice(net.random_mac_generator(mac), 1)
        self.assertEqual(['aa:bb:cc:a2:a2:a2'], list(generator))
        mock_rnd.assert_called_with(8)

    @mock.patch.object(random, 'getrandbits', return_value=0xa2)
    def test_respected_early_zeroes_generated_mac(self, mock_rnd):
        mac1 = ['00', 'bb', 'cc', '00', 'ee', 'ff']

        generator = itertools.islice(net.random_mac_generator(mac1), 1)
        self.assertEqual(['00:bb:cc:a2:a2:a2'], list(generator))

        mac2 = ['aa', '00', 'cc', '00', 'ee', 'ff']
        generator = itertools.islice(net.random_mac_generator(mac2), 1)
        self.assertEqual(['aa:00:cc:a2:a2:a2'], list(generator))

        mac3 = ['aa', 'bb', '00', '00', 'ee', 'ff']
        generator = itertools.islice(net.random_mac_generator(mac3), 1)
        self.assertEqual(['aa:bb:00:a2:a2:a2'], list(generator))
        mock_rnd.assert_called_with(8)

    @mock.patch.object(random, 'getrandbits', return_value=0xa2)
    def test_short_supplied_mac(self, mock_rnd):
        mac_base = '12:34:56:78'
        mac = mac_base.split(':')
        generator = itertools.islice(net.random_mac_generator(mac), 1)
        self.assertEqual(['12:34:56:78:a2:a2'], list(generator))
        mock_rnd.assert_called_with(8)


class TestPortDeviceOwner(base.BaseTestCase):

    def test_is_port_trusted(self):
        self.assertTrue(net.is_port_trusted(
            {'device_owner':
             constants.DEVICE_OWNER_NETWORK_PREFIX + 'dev'}))

    def test_is_port_not_trusted(self):
        self.assertFalse(net.is_port_trusted(
            {'device_owner': constants.DEVICE_OWNER_COMPUTE_PREFIX + 'dev'}))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing


def cpu_count():
    """Get the system CPU count.

    :returns: The number of CPUs on the system as an int. If there's an issue
        fetching the CPU count on the system, a default value of 1 is returned.
    """
    try:
        return multiprocessing.cpu_count()
    except NotImplementedError:
        return 1



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import re

import testtools

from neutron_lib.tests import _base as base
from neutron_lib.utils import helpers


class TestParseMappings(base.BaseTestCase):

    def parse(self, mapping_list, unique_values=True, unique_keys=True):
        return helpers.parse_mappings(mapping_list, unique_values, unique_keys)

    def test_parse_mappings_fails_for_missing_separator(self):
        with testtools.ExpectedException(ValueError):
            self.parse(['key'])

    def test_parse_mappings_fails_for_missing_key(self):
        with testtools.ExpectedException(ValueError):
            self.parse([':val'])

    def test_parse_mappings_fails_for_missing_value(self):
        with testtools.ExpectedException(ValueError):
            self.parse(['key:'])

    def test_parse_mappings_fails_for_extra_separator(self):
        with testtools.ExpectedException(ValueError):
            self.parse(['key:val:junk'])

    def test_parse_mappings_fails_for_duplicate_key(self):
        with testtools.ExpectedException(ValueError):
            self.parse(['key:val1', 'key:val2'])

    def test_parse_mappings_fails_for_duplicate_value(self):
        with testtools.ExpectedException(ValueError):
            self.parse(['key1:val', 'key2:val'])

    def test_parse_mappings_succeeds_for_one_mapping(self):
        self.assertEqual({'key': 'val'}, self.parse(['key:val']))

    def test_parse_mappings_succeeds_for_n_mappings(self):
        self.assertEqual({'key1': 'val1', 'key2': 'val2'},
                         self.parse(['key1:val1', 'key2:val2']))

    def test_parse_mappings_succeeds_for_duplicate_value(self):
        self.assertEqual({'key1': 'val', 'key2': 'val'},
                         self.parse(['key1:val', 'key2:val'], False))

    def test_parse_mappings_succeeds_for_no_mappings(self):
        self.assertEqual({}, self.parse(['']))

    def test_parse_mappings_succeeds_for_nonuniq_key(self):
        self.assertEqual({'key': ['val1', 'val2']},
                         self.parse(['key:val1', 'key:val2', 'key:val2'],
                                    unique_keys=False))

    def test_parse_mappings_succeeds_for_nonuniq_key_duplicate_value(self):
        self.assertEqual({'key': ['val']},
                         self.parse(['key:val', 'key:val', 'key:val'],
                                    unique_keys=False))


class TestCompareElements(base.BaseTestCase):

    def test_compare_elements(self):
        self.assertFalse(helpers.compare_elements([], ['napoli']))
        self.assertFalse(helpers.compare_elements(None, ['napoli']))
        self.assertFalse(helpers.compare_elements(['napoli'], []))
        self.assertFalse(helpers.compare_elements(['napoli'], None))
        self.assertFalse(helpers.compare_elements(['napoli', 'juve'],
                                                  ['juve']))
        self.assertTrue(helpers.compare_elements(['napoli', 'juve'],
                                                 ['napoli', 'juve']))
        self.assertTrue(helpers.compare_elements(['napoli', 'juve'],
                                                 ['juve', 'napoli']))


class TestDictUtils(base.BaseTestCase):

    def test_dict2str(self):
        dic = {"key1": "value1", "key2": "value2", "key3": "value3"}
        expected = "key1=value1,key2=value2,key3=value3"
        self.assertEqual(expected, helpers.dict2str(dic))

    def test_str2dict(self):
        string = "key1=value1,key2=value2,key3=value3"
        expected = {"key1": "value1", "key2": "value2", "key3": "value3"}
        self.assertEqual(expected, helpers.str2dict(string))

    def test_dict_str_conversion(self):
        dic = {"key1": "value1", "key2": "value2"}
        self.assertEqual(dic, helpers.str2dict(helpers.dict2str(dic)))

    def test_diff_list_of_dict(self):
        old_list = [{"key1": "value1"},
                    {"key2": "value2"},
                    {"key3": "value3"}]
        new_list = [{"key1": "value1"},
                    {"key2": "value2"},
                    {"key4": "value4"}]
        added, removed = helpers.diff_list_of_dict(old_list, new_list)
        self.assertEqual(added, [dict(key4="value4")])
        self.assertEqual(removed, [dict(key3="value3")])


class TestDict2Tuples(base.BaseTestCase):

    def test_dict(self):
        input_dict = {'foo': 'bar', '42': 'baz', 'aaa': 'zzz'}
        expected = (('42', 'baz'), ('aaa', 'zzz'), ('foo', 'bar'))
        output_tuple = helpers.dict2tuple(input_dict)
        self.assertEqual(expected, output_tuple)


class TestCamelize(base.BaseTestCase):

    def test_camelize(self):
        data = {'bandwidth_limit': 'BandwidthLimit',
                'test': 'Test',
                'some__more__dashes': 'SomeMoreDashes',
                'a_penguin_walks_into_a_bar': 'APenguinWalksIntoABar'}

        for s, expected in data.items():
            self.assertEqual(expected, helpers.camelize(s))


class TestRoundVal(base.BaseTestCase):

    def test_round_val_ok(self):
        for expected, value in ((0, 0),
                                (0, 0.1),
                                (1, 0.5),
                                (1, 1.49),
                                (2, 1.5)):
            self.assertEqual(expected, helpers.round_val(value))


class TestGetRandomString(base.BaseTestCase):

    def test_get_random_string(self):
        length = 127
        random_string = helpers.get_random_string(length)
        self.assertEqual(length, len(random_string))
        regex = re.compile('^[0-9a-fA-F]+$')
        self.assertIsNotNone(regex.match(random_string))


class TestSafeDecodeUtf8(base.BaseTestCase):

    def test_py3_decoded_valid_bytes(self):
        s = bytes('test-py2', 'utf-8')
        decoded_str = helpers.safe_decode_utf8(s)
        self.assertIsInstance(decoded_str, str)
        self.assertEqual(s, decoded_str.encode('utf-8'))

    def test_py3_decoded_invalid_bytes(self):
        s = bytes('test-py2', 'utf_16')
        decoded_str = helpers.safe_decode_utf8(s)
        self.assertIsInstance(decoded_str, str)


class TestSafeSortKey(base.BaseTestCase):

    def test_safe_sort_key(self):
        data1 = {'k1': 'v1',
                 'k2': 'v2'}
        data2 = {'k2': 'v2',
                 'k1': 'v1'}
        self.assertEqual(helpers.safe_sort_key(data1),
                         helpers.safe_sort_key(data2))

    def _create_dict_from_list(self, list_data):
        d = collections.defaultdict(list)
        for k, v in list_data:
            d[k].append(v)
        return d

    def test_safe_sort_key_mapping_ne(self):
        list1 = [('yellow', 1), ('blue', 2), ('yellow', 3),
                 ('blue', 4), ('red', 1)]
        data1 = self._create_dict_from_list(list1)
        list2 = [('yellow', 3), ('blue', 4), ('yellow', 1),
                 ('blue', 2), ('red', 1)]
        data2 = self._create_dict_from_list(list2)
        self.assertNotEqual(helpers.safe_sort_key(data1),
                            helpers.safe_sort_key(data2))

    def test_safe_sort_key_mapping(self):
        list1 = [('yellow', 1), ('blue', 2), ('red', 1)]
        data1 = self._create_dict_from_list(list1)
        list2 = [('blue', 2), ('red', 1), ('yellow', 1)]
        data2 = self._create_dict_from_list(list2)
        self.assertEqual(helpers.safe_sort_key(data1),
                         helpers.safe_sort_key(data2))



# Copyright (C) 2018 AT&T
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from neutron_lib.api.definitions import taas as taas_api_def

ALIAS = 'taas-vlan-filter'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = "VLAN filtering for Neutron Tap as a Service"
DESCRIPTION = "Neutron Tap as a Service Extension for filtering VLANs."
UPDATED_TIMESTAMP = "2018-01-14T10:00:00-00:00"
API_PREFIX = '/taas'
RESOURCE_NAME = 'tap_service'
COLLECTION_NAME = 'tap_services'
TAP_FLOW = 'tap_flow'
TAP_FLOWS = 'tap_flows'

# Regex for a comma-seperate list of integer values (VLANs)
# For ex. "9,18,27-36,45-54" or "0-4095" or "9,18,27,36"
RANGE_REGEX = r"^([0-9]+(-[0-9]+)?)(,([0-9]+(-[0-9]+)?))*$"

RESOURCE_ATTRIBUTE_MAP = {
    taas_api_def.TAP_FLOWS: {
        'vlan_filter': {'allow_post': True, 'allow_put': False,
                        'validate': {'type:regex_or_none': RANGE_REGEX},
                        'is_visible': True, 'default': None}
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = [taas_api_def.ALIAS]
OPTIONAL_EXTENSIONS = []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import qos
from neutron_lib.services.qos import constants as q_const
from neutron_lib.tests.unit.api.definitions import base


class QoSDefinitionTestCase(base.DefinitionBaseTestCase):
    extension_module = qos
    extension_resources = (qos.POLICIES, qos.RULE_TYPES)
    extension_subresources = (qos.BANDWIDTH_LIMIT_RULES,
                              qos.DSCP_MARKING_RULES,
                              qos.MIN_BANDWIDTH_RULES)
    extension_attributes = (q_const.DIRECTION, q_const.MAX_BURST, 'type',
                            q_const.DSCP_MARK, q_const.MIN_KBPS, 'rules',
                            q_const.MAX_KBPS, q_const.QOS_POLICY_ID)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# neutron-fwaas constants
from neutron_lib import constants

FIREWALL_GROUPS = 'firewall_groups'
FIREWALL_POLICIES = 'firewall_policies'
FIREWALL_RULES = 'firewall_rules'
FIREWALLS = 'firewalls'

FWAAS_ALLOW = "allow"
FWAAS_DENY = "deny"
FWAAS_REJECT = "reject"
FW_VALID_ACTION_VALUES = [FWAAS_ALLOW, FWAAS_DENY, FWAAS_REJECT]

# Firewall Protocol List

FW_PROTOCOL_VALUES = list(constants.IPTABLES_PROTOCOL_MAP.keys()) + [None]

# a default resource, such as auto allocated topology is_default network
IS_DEFAULT = 'is_default'



# Copyright (c) 2019 OpenStack Foundation
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import converters
from neutron_lib.api.definitions import l3
from neutron_lib import constants
from neutron_lib.db import constants as db_const

PROTOCOLS = constants.IPTABLES_PROTOCOL_MAP.keys()

# The alias of the extension.
ALIAS = 'l3-conntrack-helper'

# Whether or not this extension is simply signaling behavior to the user
# or it actively modifies the attribute map.
IS_SHIM_EXTENSION = False

# Whether the extension is marking the adoption of standardattr model for
# legacy resources, or introducing new standardattr attributes. False or
# None if the standardattr model is adopted since the introduction of
# resource extension.
# If this is True, the alias for the extension should be prefixed with
# 'standard-attr-'.
IS_STANDARD_ATTR_EXTENSION = False

# The name of the extension.
NAME = 'L3 Conntrack helper'

# A prefix for API resources. An empty prefix means that the API is going
# to be exposed at the v2/ level as any other core resource.
API_PREFIX = '/' + l3.ROUTERS

# The description of the extension.
DESCRIPTION = "Allow adding CT target rules for conntrack helper"

# A timestamp of when the extension was introduced.
UPDATED_TIMESTAMP = "2019-04-04T10:00:00-00:00"

# The name of the resource.
RESOURCE_NAME = 'conntrack_helper'

# The plural for the resource.
COLLECTION_NAME = 'conntrack_helpers'

# parent
PARENT_RESOURCE_NAME = l3.ROUTER
PARENT_COLLECTION_NAME = l3.ROUTERS

ID = 'id'
PROJECT_ID = 'project_id'
PROTOCOL = 'protocol'
PORT = 'port'
HELPER = 'helper'
RESOURCE_ATTRIBUTE_MAP = {}

# The subresource attribute map for the extension. It adds child resources
# to main extension's resource. The subresource map must have a parent and
# a parameters entry. If an extension does not need such a map, None can
# be specified (mandatory).
SUB_RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'parent': {'collection_name': PARENT_COLLECTION_NAME,
                   'member_name': PARENT_RESOURCE_NAME},
        'parameters': {
            ID: {'allow_post': False,
                 'allow_put': False,
                 'validate': {'type:uuid': None},
                 'is_visible': True,
                 'primary_key': True,
                 'is_sort_key': True,
                 'is_filter': True},
            PROJECT_ID: {'allow_post': True,
                         'allow_put': False,
                         'validate': {
                             'type:string': db_const.PROJECT_ID_FIELD_SIZE},
                         'required_by_policy': True,
                         'is_visible': False},
            PROTOCOL: {'allow_post': True, 'allow_put': True,
                       'validate': {'type:values': PROTOCOLS},
                       'is_visible': True,
                       'convert_to': converters.convert_to_protocol,
                       'is_sort_key': True,
                       'is_filter': True},
            PORT: {'allow_post': True, 'allow_put': True,
                   'convert_to': converters.convert_to_int,
                   'validate': {'type:range': [1, 65535]},
                   'is_visible': True,
                   'is_sort_key': True,
                   'is_filter': True},
            HELPER: {'allow_post': True, 'allow_put': True,
                     'convert_to': converters.convert_to_string,
                     'validate': {'type:string': 64},
                     'is_visible': True,
                     'is_sort_key': True,
                     'is_filter': True},
        }
    }
}

# The action map: it associates verbs with methods to be performed on
# the API resource.
ACTION_MAP = {
}

# The action status.
ACTION_STATUS = {
}

# The list of required extensions.
REQUIRED_EXTENSIONS = [l3.ALIAS]

# The list of optional extensions.
OPTIONAL_EXTENSIONS = [
]



# Copyright (c) 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port
from neutron_lib.db import constants


ALIAS = 'port-device-profile'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Port device profile'
DESCRIPTION = "Expose the port device profile (Cyborg)"
UPDATED_TIMESTAMP = "2020-12-17T10:00:00-00:00"
RESOURCE_NAME = port.RESOURCE_NAME
COLLECTION_NAME = port.COLLECTION_NAME
DEVICE_PROFILE = 'device_profile'

RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        DEVICE_PROFILE: {
            'allow_post': True,
            'allow_put': False,
            'validate': {
                'type:string_or_none': constants.DESCRIPTION_FIELD_SIZE},
            'default': None,
            'is_visible': True}
    },
}

SUB_RESOURCE_ATTRIBUTE_MAP = None
ACTION_MAP = {}
ACTION_STATUS = {}
REQUIRED_EXTENSIONS = []
OPTIONAL_EXTENSIONS = []



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from webob import exc

from neutron_lib.api.validators import allowedaddresspairs as validator
from neutron_lib.exceptions import allowedaddresspairs as addr_exc
from neutron_lib.tests import _base as base


class TestAllowedAddressPairs(base.BaseTestCase):

    def test__validate_allowed_address_pairs_not_a_list(self):
        for d in [{}, set(), 'abc', True, 1]:
            self.assertRaisesRegex(
                exc.HTTPBadRequest, 'must be a list',
                validator._validate_allowed_address_pairs, d)

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_exhausted(self, mock_conf):
        mock_conf.max_allowed_address_pair = 1
        self.assertRaises(
            addr_exc.AllowedAddressPairExhausted,
            validator._validate_allowed_address_pairs,
            [{}, {}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_invalid_mac(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaisesRegex(
            exc.HTTPBadRequest, 'is not a valid MAC address',
            validator._validate_allowed_address_pairs,
            [{'mac_address': 1}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_missing_ip(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaises(
            addr_exc.AllowedAddressPairsMissingIP,
            validator._validate_allowed_address_pairs,
            [{'ip_adress': '192.168.1.11'}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_duplicate(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaises(
            addr_exc.DuplicateAddressPairInRequest,
            validator._validate_allowed_address_pairs,
            [{'ip_address': '192.168.1.11'},
             {'ip_address': '192.168.1.11'}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_invalid_attrs(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaisesRegex(
            exc.HTTPBadRequest, 'Unrecognized attribute',
            validator._validate_allowed_address_pairs,
            [{'ip_address': '192.168.1.11'},
             {'ip_address': '192.168.1.12', 'idk': True}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_invalid_subnet(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaisesRegex(
            exc.HTTPBadRequest, 'is not a valid IP subnet',
            validator._validate_allowed_address_pairs,
            [{'ip_address': '192.168.1.11'},
             {'ip_address': '192.168.1.0/a'}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_invalid_ip_address(
            self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertRaisesRegex(
            exc.HTTPBadRequest, 'is not a valid IP address',
            validator._validate_allowed_address_pairs,
            [{'ip_address': '192.168.1.a'},
             {'ip_address': '192.168.1.2'}])

    @mock.patch.object(cfg, 'CONF')
    def test__validate_allowed_address_pairs_good_data(self, mock_conf):
        mock_conf.max_allowed_address_pair = 3
        self.assertIsNone(validator._validate_allowed_address_pairs(
            [{'ip_address': '192.1.1.11'}, {'ip_address': '19.1.1.11'}]))



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib._i18n import _
from neutron_lib import exceptions


class DNSDomainNotFound(exceptions.NotFound):
    message = _("Domain %(dns_domain)s not found in the external DNS service")


class DuplicateRecordSet(exceptions.Conflict):
    message = _("Name %(dns_name)s is duplicated in the external DNS service")


class ExternalDNSDriverNotFound(exceptions.NotFound):
    message = _("External DNS driver %(driver)s could not be found.")


class InvalidPTRZoneConfiguration(exceptions.Conflict):
    message = _("Value of %(parameter)s has to be multiple of %(number)s, "
                "with maximum value of %(maximum)s and minimum value of "
                "%(minimum)s")


class ExternalDNSOverQuota(exceptions.OverQuota):
    message = _("External DNS Quota exceeded for resources: %(resource)s.")



# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging as std_logging
import os
import os.path
import random
from unittest import mock

import fixtures
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_utils import strutils
import pbr.version
import testtools

from neutron_lib._i18n import _
from neutron_lib import constants
from neutron_lib import exceptions
from neutron_lib import fixture

from neutron_lib.tests import _post_mortem_debug as post_mortem_debug


CONF = cfg.CONF
LOG_FORMAT = "%(asctime)s %(levelname)8s [%(name)s] %(message)s"

ROOTDIR = os.path.dirname(__file__)
ETCDIR = os.path.join(ROOTDIR, 'etc')


def etcdir(*p):
    return os.path.join(ETCDIR, *p)


def fake_use_fatal_exceptions(*args):
    return True


def get_rand_name(max_length=None, prefix='test'):
    """Return a random string.

    The string will start with 'prefix' and will be exactly 'max_length'.
    If 'max_length' is None, then exactly 8 random characters, each
    hexadecimal, will be added. In case len(prefix) <= len(max_length),
    ValueError will be raised to indicate the problem.
    """

    if max_length:
        length = max_length - len(prefix)
        if length <= 0:
            raise ValueError("'max_length' must be bigger than 'len(prefix)'.")

        suffix = ''.join(str(random.randint(0, 9)) for i in range(length))
    else:
        suffix = hex(random.randint(0x10000000, 0x7fffffff))[2:]
    return prefix + suffix


def get_rand_device_name(prefix='test'):
    return get_rand_name(
        max_length=constants.DEVICE_NAME_MAX_LEN, prefix=prefix)


def bool_from_env(key, strict=False, default=False):
    value = os.environ.get(key)
    return strutils.bool_from_string(value, strict=strict, default=default)


def get_test_timeout(default=0):
    return int(os.environ.get('OS_TEST_TIMEOUT', default))


def sanitize_log_path(path):
    # Sanitize the string so that its log path is shell friendly
    return path.replace(' ', '-').replace('(', '_').replace(')', '_')


class AttributeDict(dict):

    """Provide attribute access (dict.key) to dictionary values."""

    def __getattr__(self, name):
        """Allow attribute access for all keys in the dict."""
        if name in self:
            return self[name]
        raise AttributeError(_("Unknown attribute '%s'.") % name)


class BaseTestCase(testtools.TestCase):

    @staticmethod
    def config_parse(conf=None, args=None):
        """Create the default configurations."""
        if args is None:
            args = []
        args += ['--config-file', etcdir('neutron_lib.conf')]
        if conf is None:
            version_info = pbr.version.VersionInfo('neutron-lib')
            cfg.CONF(args=args, project='neutron_lib',
                     version='%%(prog)s %s' % version_info.release_string())
        else:
            conf(args)

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(fixture.PluginDirectoryFixture())

        # Enabling 'use_fatal_exceptions' allows us to catch string
        # substitution format errors in exception messages.
        mock.patch.object(exceptions.NeutronException, 'use_fatal_exceptions',
                          return_value=True).start()

        db_options.set_defaults(cfg.CONF, connection='sqlite://')

        self.useFixture(fixtures.MonkeyPatch(
            'oslo_config.cfg.find_config_files',
            lambda project=None, prog=None, extension=None: []))

        self.setup_config()

        # Configure this first to ensure pm debugging support for setUp()
        debugger = os.environ.get('OS_POST_MORTEM_DEBUGGER')
        if debugger:
            self.addOnException(post_mortem_debug.get_exception_handler(
                debugger))

        # Make sure we see all relevant deprecation warnings when running tests
        self.useFixture(fixture.WarningsFixture())

        if bool_from_env('OS_DEBUG'):
            _level = std_logging.DEBUG
        else:
            _level = std_logging.INFO
        capture_logs = bool_from_env('OS_LOG_CAPTURE')
        if not capture_logs:
            std_logging.basicConfig(format=LOG_FORMAT, level=_level)
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(
                format=LOG_FORMAT,
                level=_level,
                nuke_handlers=capture_logs,
            ))

        test_timeout = get_test_timeout()
        if test_timeout == -1:
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        # If someone does use tempfile directly, ensure that it's cleaned up
        self.useFixture(fixtures.NestedTempfile())
        self.useFixture(fixtures.TempHomeDir())

        self.addCleanup(mock.patch.stopall)

        if bool_from_env('OS_STDOUT_CAPTURE'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if bool_from_env('OS_STDERR_CAPTURE'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

        self.addOnException(self.check_for_systemexit)
        self.orig_pid = os.getpid()

    def get_new_temp_dir(self):
        """Create a new temporary directory.

        :returns fixtures.TempDir
        """
        return self.useFixture(fixtures.TempDir())

    def get_default_temp_dir(self):
        """Create a default temporary directory.

        Returns the same directory during the whole test case.

        :returns fixtures.TempDir
        """
        if not hasattr(self, '_temp_dir'):
            self._temp_dir = self.get_new_temp_dir()
        return self._temp_dir

    def check_for_systemexit(self, exc_info):
        if isinstance(exc_info[1], SystemExit):
            if os.getpid() != self.orig_pid:
                # Subprocess - let it just exit
                raise
            # This makes sys.exit(0) still a failure
            self.force_failure = True

    def assertOrderedEqual(self, expected, actual):
        expect_val = self.sort_dict_lists(expected)
        actual_val = self.sort_dict_lists(actual)
        self.assertEqual(expect_val, actual_val)

    def sort_dict_lists(self, dic):
        for key, value in dic.items():
            if isinstance(value, list):
                dic[key] = sorted(value)
            elif isinstance(value, dict):
                dic[key] = self.sort_dict_lists(value)
        return dic

    def assertDictSupersetOf(self, expected_subset, actual_superset):
        """Checks that actual dict contains the expected dict.

        After checking that the arguments are of the right type, this checks
        that each item in expected_subset is in, and matches, what is in
        actual_superset. Separate tests are done, so that detailed info can
        be reported upon failure.
        """
        if not isinstance(expected_subset, dict):
            self.fail("expected_subset (%s) is not an instance of dict" %
                      type(expected_subset))
        if not isinstance(actual_superset, dict):
            self.fail("actual_superset (%s) is not an instance of dict" %
                      type(actual_superset))
        for k, v in expected_subset.items():
            self.assertIn(k, actual_superset)
            self.assertEqual(v, actual_superset[k],
                             "Key %(key)s expected: %(exp)r, actual %(act)r" %
                             {'key': k, 'exp': v, 'act': actual_superset[k]})

    def setup_config(self, args=None):
        """Tests that need a non-default config can override this method."""
        self.config_parse(args=args)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib.api.definitions import agent
from neutron_lib.api import validators
from neutron_lib.api.validators import availability_zone as az_validator


validators.add_validator('availability_zone_hint_list',
                         az_validator._validate_availability_zone_hints)

AZ_HINTS = 'availability_zone_hints'

ALIAS = 'availability_zone'
IS_SHIM_EXTENSION = False
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Availability Zone'
API_PREFIX = ''
DESCRIPTION = 'The availability zone extension.'
UPDATED_TIMESTAMP = '2015-01-01T10:00:00-00:00'
RESOURCE_NAME = 'availability_zone'
COLLECTION_NAME = RESOURCE_NAME + 's'
RESOURCE_ATTRIBUTE_MAP = {
    COLLECTION_NAME: {
        'name': {'is_visible': True, 'is_filter': True},
        'resource': {'is_visible': True, 'is_filter': True},
        'state': {'is_visible': True, 'is_filter': True}
    },
    agent.COLLECTION_NAME: {
        RESOURCE_NAME: {
            'allow_post': False, 'allow_put': False, 'is_filter': True,
            'is_visible': True}
    }
}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [agent.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}
